module Draw(
  Point(..),
  pt, smult, fromPolar,
  Draw,
  initEnv,
  line, rect, polygon,
  translate, rotate, scale,
  getColour, setColour, nextColour,
  render, draw
) where

import Data.List(intercalate)
import qualified Control.Monad.Trans.State.Lazy as S

import Text.Blaze((!), stringValue) 
-- import Text.Blaze.Svg11(mkPath, l, m)
import qualified Text.Blaze.Svg11 as S
import qualified Text.Blaze.Svg11.Attributes as A

-- import Text.Blaze.Renderer.Utf8 (renderMarkup)
import Text.Blaze.Renderer.String (renderMarkup)

--
-- 0. Points
--

data Point = Pt { ptx :: Double, pty :: Double } deriving Eq

-- | Convenience function: create a 'Point' from given coordinates.
pt :: (Double, Double) -> Point
pt (x, y) = Pt x y

instance Num Point where
  Pt x1 y1 + Pt x2 y2 = Pt (x1+ x2) (y1+ y2)
  Pt x1 y1 * Pt x2 y2 = undefined -- Pt (x1* x2) (y1* y2)
  abs (Pt x y)        = Pt (abs x) (abs y)
  signum (Pt x y)     = Pt (signum x) (signum y)
  negate (Pt x y)     = Pt (-x) (-y)
  fromInteger i       = Pt (fromInteger i) 0

instance Show Point where
  show (Pt x y) = "("++ show x++ ", "++ show y++ ")"  

-- | Scalar multiplication
smult :: Double-> Point-> Point
smult r (Pt x y) = Pt (r* x) (r* y)

-- | Construction from polar form
fromPolar :: Double-> Double-> Point
fromPolar r phi = Pt (r* cos phi) (r* sin phi)

--
-- 1. The Draw monad
--

type Colour = (Int, Int, Int)

data Env  = Env { strkWdth :: Int
                , colour   :: Colour
                , graphics :: S.Svg
                }

-- The Draw monad: a state monad with the environment
type Draw  = S.State Env 

-- Accessing the environment
set :: (Env-> Env)-> Draw ()
set f = S.state $ \e-> ((), f e)

get :: (Env-> a)-> Draw a
get f = S.state $ \e-> (f e, e)

--
-- 2. Primitive Drawing commands
--
line :: Point-> Point-> Draw ()
line p q = 
  addWithAttrs $ S.line ! A.x1 (dVal $ ptx p) ! A.y1 (dVal $ pty p)
                       ! A.x2 (dVal $ ptx q) ! A.y2 (dVal $ pty q)

rect :: Point-> Point-> Draw ()
rect (Pt px py) (Pt qx qy) = polygon [Pt px py, Pt qx py, Pt qx qy, Pt px qy]

polygon :: [Point]-> Draw ()
polygon ps = addWithAttrs $ S.polygon ! A.points (stringValue pts) where
   pts = intercalate " " $ map (\p-> (showd $ ptx p) ++ ","++ (showd $ pty p)) ps

 
--
-- 3. Transformations
--

transform :: S.AttributeValue-> Draw()-> Draw()
transform av d = do
  drw <- get $ \e-> graphics $ S.execState d (e {graphics= return ()})
  addSVG $ S.g ! A.transform av $ drw
  
rotate :: Point-> Double-> Draw ()-> Draw ()
rotate p phi = transform (S.rotateAround phi (ptx p) (pty p))

scale :: Double-> Draw()-> Draw ()
scale r = transform (S.scale r r)

translate :: Point -> Draw () -> Draw ()
translate p = transform (S.translate (ptx p) (pty p))


--
-- 4. SVG Attributes -- colours and such
-- 

dVal :: Double-> S.AttributeValue
dVal = stringValue. showd

showd :: Double-> String
showd = (show :: Int-> String) . round

intVal :: Int-> S.AttributeValue
intVal = stringValue . show

rgbVal :: Colour-> S.AttributeValue
rgbVal (r, g, b) =
  stringValue $ "rgb("++ show r++ ","++ show g++ ","++ show b++ ")"

withAttrs :: S.Svg-> Env-> S.Svg
withAttrs s e = s ! A.strokeWidth (intVal $ strkWdth e)
                  ! A.stroke (rgbVal $ colour e)
                  ! A.fill (rgbVal $ colour e)

addWithAttrs :: S.Svg-> Draw () 
addWithAttrs s1 = do s2<- get $ withAttrs s1; addSVG s2

addSVG :: S.Svg-> Draw ()
addSVG s = set $ \e-> e { graphics = do graphics e; s}

-- Colours
setColour :: Int-> Int-> Int-> Draw ()
setColour r g b =
  set $ \e-> e{ colour= (adj r, adj g, adj b) } where adj c = c `mod` 256

getColour :: Draw Colour
getColour = get colour

nextCol :: Colour-> Colour
nextCol (r, g, b) = (r+ 7, g+ 17, b+ 23)

nextColour :: Draw ()
nextColour = do (r, g, b)<- getColour >>= return . nextCol
                setColour r g b

-- Stroke width
strokeWidth :: Int-> Draw ()
strokeWidth d = set $ \e-> e {strkWdth = d }

--
-- 5. Execution--- running a drawing computation
--

initEnv :: Env
initEnv = Env { strkWdth = 1
              , colour= (0, 0, 0)
              , graphics = return ()
              }
              
render :: Int-> Int-> Draw()-> S.Svg
render w h d =
  S.svg ! A.width (stringValue $ show (2*w))
        ! A.height (stringValue $ show (2*h))
    $ S.g ! A.transform (S.translate w h)
    $ S.evalState (d >> get graphics) initEnv
   
draw :: String-> S.Svg -> IO ()
draw fnm d = do
   writeFile fnm $ renderMarkup $ d
   putStrLn $ "Written to file "++ fnm

